import abc
import unittest
from unittest.mock import patch

class TestGCPPubsub(unittest.TestCase):

    @patch('dotenv.load_dotenv')
    def test_import(self, load):

        load.side_effect = [None] * 3

        from tesselite.pubsub import GCPPubSub

        assert isinstance(GCPPubSub, abc.ABCMeta)

    def test_publish(self):
        assert True  # add assertion here

    def test_consume(self):
        assert True  # add assertion here

if __name__ == '__main__':
    unittest.main()



import os

from tesselite import App

def load(var):
    try:
        return os.environ[var]
    except KeyError:
         App.Logger.fatal(f"the env variable '{var}' is missing.")
         exit(1)


class RedisEnv:
    HOST = os.environ.get("REDIS_HOST", "localhost")
    PORT = int(os.environ.get("REDIS_PORT", "6379"))
    DB = int(os.environ.get("REDIS_DB", "0"))
    PASSWORD = os.environ.get("REDIS_PASSWORD", "tesselite")
    TOPIC_NAME = 'tesselite-pubsub'
    SUBSCRIPTION_NAME = 'tesselite'


class GCPEnv:
    from dotenv import load_dotenv
    load_dotenv()
    GOOGLE_PROJECT = load('GOOGLE_PROJECT')
    GOOGLE_APPLICATION_CREDENTIALS = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')
    TOPIC_NAME = 'tesselite-pubsub'
    SUBSCRIPTION_NAME = 'tesselite'

class RabbitMQEnv:
    from dotenv import load_dotenv
    load_dotenv()
    HOST = os.environ.get("RABBITMQ_HOST", "localhost")
    PORT = int(os.environ.get("RABBITMQ_PORT", "5672"))
    USER = os.environ.get("RABBITMQ_USER", "tesselite")
    PASSWORD = os.environ.get("RABBITMQ_PASSWORD", "tesselite")
    TOPIC_NAME = 'tesselite-pubsub'
    SUBSCRIPTION_NAME = 'tesselite'
    VHOST = '/'






# High-Level Setters
import os

from dotenv import load_dotenv

from tesselite.exceptions import ConfigurationException
from tesselite.logger import Logger

load_dotenv()

class App:
    Name = "tesselite"
    BrokerType = os.environ.get("BROKER", "GCP-PUBSUB")
    Logger = Logger(Name) # root logger





"""
Handle messaging (publish/consume)
"""
import abc
from typing import Union
import redis
from typing import Callable
import pika

from google.api_core import exceptions as google_api_core_exceptions
from google.cloud.pubsub_v1 import PublisherClient
from google.cloud.pubsub_v1 import SubscriberClient
import socket

from pika.exceptions import ProbableAuthenticationError, StreamLostError, IncompatibleProtocolError, \
    AMQPConnectionError, ConnectionWrongStateError

from tesselite import Logger, App
from tesselite.exceptions import connexion
from tesselite.exceptions import MessageProcessingException


class Pubsub(abc.ABC):
    """
    Publish messages into a Broker
    """

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @abc.abstractmethod
    def open(self):
        pass

    @abc.abstractmethod
    def close(self):
        pass

    @abc.abstractmethod
    def publish(self, msg: str):
        raise NotImplementedError()

    @abc.abstractmethod
    def consume(self, callback: Callable, deadLetter: str, **kwargs):
        raise NotImplementedError()


class RedisPubsub(Pubsub):

    def __init__(self, topic: str, log_name: str = "redis-pubsub"):
        from tesselite.config import RedisEnv
        self.logger = Logger(log_name)
        self.log_name = log_name
        self.logger.debug("loading ..")
        self._pubsub = None
        self._env = RedisEnv()
        self._client = None
        self._topic = topic if topic else self._env.TOPIC_NAME
        self.logger.debug("\n"
                          f"\tBROKER: REDIS\n"
                          f"\tHOST: {self._env.HOST}\n"
                          f"\tPORT: {self._env.PORT}\n"
                          f"\tPASSWORD: "
                          f"( {'hidden' if self._env.PASSWORD else 'empty'} )")

    @property
    def topic(self):
        return self._topic

    @connexion(expected_errors=(socket.gaierror, redis.exceptions.ConnectionError,))
    def open(self):
        self.logger.debug("connecting ..")
        self._client = redis.Redis(host=self._env.HOST, port=self._env.PORT,
                                   db=self._env.DB, password=self._env.PASSWORD)
        self.logger.debug("pinging ..")
        self._client.ping()
        self.logger.info("ready.")

    def close(self):
        self._client.close()
        self.logger.info("terminated.")

    # if standard networks errors, backoff the loop
    @connexion(expected_errors=(socket.gaierror, redis.exceptions.ConnectionError,))
    def publish(self, msg: str):
        self._client.publish(self._topic, msg)

    # consume loop
    # if standard networks errors, backoff the loop
    @connexion(expected_errors=(socket.gaierror, redis.exceptions.ConnectionError,))
    def consume(self, callback: Callable, deadLetter: str = None, **kwargs):
        self.logger.debug("consuming ..")

        """
        callback: function called with message payload
        e.g.,
            def callback(message: str):
                print(message)
        deadLetter: a backup topic where unconsumed events are pushed
        """
        # only events happening after subscription are processed
        self._pubsub: redis.client.PubSub = self._client.pubsub(ignore_subscribe_messages=False)
        self._pubsub.subscribe(self._topic)

        # exec wrapper
        @connexion(expected_errors=(socket.gaierror, redis.exceptions.ConnectionError,))
        def exec_callback(message):
            if message['type'] == 'message':
                data = msg['data'].decode()
                callback(data)

        # event loop
        try:
            for msg in self._pubsub.listen():
                try:
                    exec_callback(msg)
                except Exception as err:
                    self.logger.error(err, stack_info=True)
                    if deadLetter:
                        self._client.publish(deadLetter, msg)
                    raise
        except KeyboardInterrupt:
            self.logger.info("graceful exit")
        except Exception:
            raise


class GCPPubSub(Pubsub):

    def __init__(self, topic: str, log_name: str = "gcp-pubsub"):
        from tesselite import Logger
        from tesselite.config import GCPEnv
        self.logger = Logger(log_name)
        self._topic_path = None
        self.logger.debug("loading ..")
        self._env = GCPEnv()
        self._topic = topic if topic else self._env.TOPIC_NAME
        self._topic_path = None
        self._subscription = self._env.SUBSCRIPTION_NAME
        self._subscription_path = None
        self._publisher_client = None
        self._subscriber_client = None
        self.logger.info("\n"
                          f"\tBROKER: GCP\n"
                          f"\tPROJECT: {self._env.GOOGLE_PROJECT}\n"
                          f"\tTOPIC: {self._topic}\n"
                          f"\tCREDENTIALS: "
                          f"( {'hidden' if self._env.GOOGLE_APPLICATION_CREDENTIALS else 'empty'} )")

    @property
    def topic(self):
        return self._topic

    def open(self):
        self._publisher_client = PublisherClient()
        # set topic location
        self._topic_path = self._publisher_client.topic_path(self._env.GOOGLE_PROJECT, self.topic)
        # check topic
        self.check_topic()
        self.logger.info("ready.")

    def close(self):
        self._publisher_client.__exit__(None, None, None)
        if self._subscriber_client is not None:
            self._subscriber_client.__exit__(None, None, None)
        self.logger.info("terminated.")

    @connexion(
        expected_errors=(google_api_core_exceptions.ServiceUnavailable, google_api_core_exceptions.Forbidden),
        noisy_errors=(google_api_core_exceptions.AlreadyExists,)
    )
    def check_topic(self):
        """
        creates a new topic if it doesn't exist
        in production, Terraform must create topics
        """
        try:
            self.logger.debug(f"checking  .. topic={self._topic_path}")
            self._publisher_client.get_topic(topic=self._topic_path, retry=None)
        except google_api_core_exceptions.NotFound:
            self.logger.info(f"creating new topic .. {self._topic_path}")
            self._publisher_client.create_topic(name=self._topic_path)
        except google_api_core_exceptions.Forbidden as err:
            self.logger.error(f"checking topic .. {self._topic_path} failed => reason:forbidden.")
            self.logger.error(err)
        except:
            raise

    @connexion(
        expected_errors=(google_api_core_exceptions.ServiceUnavailable, google_api_core_exceptions.Forbidden),
        noisy_errors=(google_api_core_exceptions.AlreadyExists,)
    )
    def check_subscription(self):
        """
        creates a new subscription if it doesn't exist
        in production, Terraform must create subscriptions for better retention
        """
        # check topic
        self.check_topic()
        try:
            self.logger.debug(f"checking subscription .. {self._subscription_path}")
            resource = self._subscriber_client.get_subscription(subscription=self._subscription_path)
            self.logger.debug(f"name={resource.name}, topic={resource.topic}")
            if resource.topic != self._topic_path:
                self.logger.fatal("\n"
                                  "\tdetected inconsistent subscription ..\n"
                                 f"\t* consumer's topic: {self._topic_path}\n"
                                 f"\t* subscription's topic: {resource.topic}")
                exit(1)
            return
        except (google_api_core_exceptions.NotFound, google_api_core_exceptions.InvalidArgument):
            self.logger.info(f"registering new subscription .. {self._subscription_path}")
            self._subscriber_client.create_subscription(name=self._subscription_path, topic=self._topic_path)
        except google_api_core_exceptions.Forbidden as err:
            self.logger.error(f"registering subscription .. {self._subscription_path} failed => reason:forbidden.")
            self.logger.error(err)
        except Exception as err:
            self.logger.error(err, stack_info=True)
            raise

    @connexion(
        expected_errors=(google_api_core_exceptions.ServiceUnavailable,),
        noisy_errors=(google_api_core_exceptions.AlreadyExists,)
    )
    def publish(self, msg: str):
        call = self._publisher_client.publish(self._topic_path, msg.encode() if isinstance(msg, str) else msg)
        return call.result()

    @connexion(expected_errors=(google_api_core_exceptions.ServiceUnavailable,
                                google_api_core_exceptions.RetryError),
               noisy_errors=(
               google_api_core_exceptions.AlreadyExists, google_api_core_exceptions.NotFound, TimeoutError)
               )
    def consume(self, callback: Callable, subscription: str = None, deadLetter: str = None):
        """
        callback: function called with message payload
        deadLetter: is a backup topic where unconsumed events are pushed
        """

        # exec wrapper
        @connexion(expected_errors=(google_api_core_exceptions.ServiceUnavailable,
                                    MessageProcessingException, google_api_core_exceptions.NotFound,))
        def exec_callback(message):
            try:
                callback(message.data.decode())
                message.ack()
            except Exception as err:
                self.logger.error(f"{type(err)} {err}")
                raise MessageProcessingException()

        # event loop
        with SubscriberClient() as self._subscriber_client:
            # check subscription
            self.logger.debug("checking ..")
            self._subscription = subscription if subscription is not None else self._env.SUBSCRIPTION_NAME
            # append topic name to create 1:1 mapping with topic
            self._subscription = f"s-{self._subscription}-t-{self._topic}"
            self._subscription_path = self._subscriber_client.subscription_path(subscription=self._subscription,
                                                                                project=self._env.GOOGLE_PROJECT)
            self.check_subscription()
            self.logger.debug("start consuming ..")
            future = self._subscriber_client.subscribe(subscription=self._subscription_path,
                                                       callback=exec_callback,
                                                       use_legacy_flow_control=True)
            try:
                future.result()
            except KeyboardInterrupt:
                future.cancel()
            except Exception:
                raise

class RabbitMQPubSub(Pubsub):
    def __init__(self, topic: str, log_name: str = "rabbitmq-pubsub"):
        from tesselite import Logger
        from tesselite.config import RabbitMQEnv
        self.logger = Logger(log_name)
        self.logger.debug("loading ..")
        self._env = RabbitMQEnv()
        self._topic = topic if topic else self._env.TOPIC_NAME
        self._connection: pika.adapters.blocking_connection.BlockingConnection = None
        self._channel: pika.adapters.blocking_connection.BlockingChannel = None
        self.logger.info("\n"
                         f"\tBROKER: RABBITMQ\n"
                         f"\tTOPIC: {self._topic}\n"
                         f"\tHOST: {self._env.HOST}\n"
                         f"\tPORT: {self._env.PORT}\n"
                         f"\tCREDENTIALS: "
                         f"( {'hidden' if self._env.PASSWORD else 'empty'} )"
        )

    @connexion(
        expected_errors=(ProbableAuthenticationError,
                         IncompatibleProtocolError,
                         AMQPConnectionError,
                         ConnectionWrongStateError),
        noisy_errors=())
    def open(self):
        credentials = pika.PlainCredentials(username=self._env.USER, password=self._env.PASSWORD)
        parameters = pika.ConnectionParameters(self._env.HOST, self._env.PORT, self._env.VHOST, credentials)
        self._connection = pika.BlockingConnection(parameters)
        self._channel = self._connection.channel()
        self._channel.exchange_declare(exchange=self._topic, exchange_type='topic')

    def close(self):
        self._channel.cancel()
        self._connection.close()


    @connexion(
        expected_errors=(ProbableAuthenticationError, IncompatibleProtocolError,
                         AMQPConnectionError, ConnectionWrongStateError),
        noisy_errors=())
    def check_subscription(self, subscription):
        queue = subscription if subscription is not None else self._env.SUBSCRIPTION_NAME
        self._channel.queue_declare(queue=queue)
        self._channel.queue_bind(queue=queue, exchange=self._topic, routing_key=queue)
        return queue

    def check_connection(self):
        if self._connection is None or self._connection.is_closed:
            self.open()

    @connexion(
        expected_errors=(ProbableAuthenticationError, IncompatibleProtocolError,
                         AMQPConnectionError, ConnectionWrongStateError),
        noisy_errors=())
    def publish(self, msg: str, subscription:str='*'):
        self.check_connection()
        self._channel.basic_publish(exchange=self._topic,
                                    routing_key=subscription,
                                    body=msg.encode())

    @connexion(
        expected_errors=(ProbableAuthenticationError, IncompatibleProtocolError,
                         AMQPConnectionError, ConnectionWrongStateError),
        noisy_errors=())
    def consume(self, callback: Callable, deadLetter: str = None, subscription='*'):
        self.check_connection()
        queue = self.check_subscription(subscription=subscription)
        for method_frame, properties, body in self._channel.consume(queue):
            callback(body.decode())
            self._channel.basic_ack(method_frame.delivery_tag)


def pubsubFactory(broker: str = App.BrokerType) -> Union[type(RedisPubsub), type(GCPPubSub)]:
    """creates a publisher object
    :type broker: str: broker backend.
        supported:
          - redis
          - gcp-pubsub
          - rabbitmq
    """
    if broker.upper() == "REDIS":
        return RedisPubsub
    elif broker.upper() == "GCP-PUBSUB":
        return GCPPubSub
    elif broker.upper() == "RABBITMQ":
        return RabbitMQPubSub
    else:
        App.Logger.fatal(f"Broker type <{broker}> not available yet.")
        exit(1)



from typing import Callable, Type

from retry import retry


def graceful(action: Callable):

    from tesselite import App

    def run(*args, **kwargs):

        try:
            return action(*args, **kwargs)
        except KeyboardInterrupt:
            App.Logger.info("bye.")
        except:
            raise

    return run


def connexion(expected_errors:tuple, noisy_errors:tuple=()):

    from tesselite import App

    def run(function:Callable):

        @retry(expected_errors+noisy_errors, delay=2, backoff=2, max_delay=20)
        def inner(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except expected_errors as err:
                App.Logger.error(f"({function.__name__}) connexion error [{err.__class__.__name__}] => backoff.")
                raise
            except noisy_errors as err:
                App.Logger.debug(f"({function.__name__}) connexion error [{err.__class__.__name__}] => backoff.")
                raise
            except KeyboardInterrupt:
                raise
            except Exception as err:
                App.Logger.error(f"({function.__name__}) unknown error => {err}")
                raise

        return inner

    return run


class MessageProcessingException(Exception):
    pass


class ConfigurationException(Exception):
    pass



import logging
import os
import sys


class Logger(logging.Logger):

    def __init__(self, name):
        super().__init__(name=name)
        self.set_handler()

    def set_handler(self):
        hdr = logging.StreamHandler(sys.stdout)
        # log format
        fmt = logging.Formatter('[%(name)s][%(levelname)s][%(asctime)s] %(message)s', datefmt='%Y-%m-%d %I:%M:%S')
        hdr.setFormatter(fmt)
        # load env vars
        from dotenv import load_dotenv
        load_dotenv()
        # set loglevel
        lvl = os.environ.get('LOGLEVEL', 'ERROR').upper()
        self.setLevel(lvl)
        self.addHandler(hdr)



import threading
from time import sleep
from typing import Callable, Type

from tesselite import App
from tesselite.exceptions import graceful
from tesselite.pubsub import pubsubFactory
from tesselite.schemas import Message


@graceful
def publish(broker: str, encoder: Callable, topic: str = None, subscription:str = None):
    """
    Pattern: publish (topicA)
    :param broker: broker backend (gcp-pubsub or redis)
    :param encoder: message generator function (produces bytes or raw string)
    :param topic:
    :return:
    """
    with pubsubFactory(broker=broker)(topic=topic, log_name="publish") as pubsub:
        for msg in encoder():
            pubsub.publish(msg, subscription=subscription)


@graceful
def consume(broker: str, callback: Callable, topic: str = None, subscription: str = None):
    """
    Pattern: consume (topicA)
    :param broker:
    :param callback:
    :param topic:
    :param subscription:
    :return:
    """
    with pubsubFactory(broker=broker)(topic=topic, log_name="consume") as pubsub:
        pubsub.consume(callback=callback, deadLetter=None,
                       subscription=subscription)


def autoConsumeTest(broker: str, callback: Callable, encoder: Callable, timeout: int = None,
                    topic: str = None, subscription: str = None) -> (threading.Thread, threading.Thread):
    """
    Pattern: publish (topicA) => consume (topicA)
    :param broker: backend
    :param callback: consume callback
    :param encoder: message create
    :param timeout:
    :param topic: topic's name
    :param subscription: subscription's name
    :return: (thread_1, thread_2) to terminate
    """
    consume_thread = threading.Thread(target=consume, args=(broker, callback, topic, subscription), daemon=True)
    publish_thread = threading.Thread(target=publish, args=(broker, encoder, topic), daemon=True)

    # fire consume
    consume_thread.start()
    # consume grace
    sleep(1)
    # fire publish
    publish_thread.start()

    consume_thread.join(timeout=timeout)
    publish_thread.join(timeout=timeout)

    return consume_thread, publish_thread


@graceful
def consumeAndPublish(broker: str, transform: Callable,
                      topic_in: str, topic_out: str, subscription: str = None,
                      schema: Type[Message] = Message, polling_interval: float = 0.01, ):
    """
    Pattern: consume (topicA) => publish (topicB)
    :param broker: backend (gcp-pubsub or redis)
    :param transform: message transformation
    :param schema: object
    :param topic_in: entry topic's name
    :param topic_out: exit topic's name
    :param subscription: subscription's name
    :param polling_interval: add message time gap to control chaos
    :return: None
    """
    if topic_in == topic_out:
        App.Logger.warning("circular consumption detected.")

    with pubsubFactory(broker=broker)(topic=topic_out, log_name="consumeAndPublish") as publisher:
        def callback(message):
            deserialized: schema = schema.deserialize(message)
            sleep(polling_interval)
            publisher.publish(transform(deserialized))

        with pubsubFactory(broker=broker)(topic=topic_in, log_name="consumeAndPublish") as consumer:
            consumer.consume(callback=callback, deadLetter=None,
                             subscription=subscription)


def consumePublishThread(broker: str, transform: Callable,
                         topic_in: str, topic_out: str, subscription: str = None,
                         schema: Type[Message] = Message, polling_interval:float=0.01) -> threading.Thread:
    """
    Pattern: consume (topicA) => publish (topicB)
    :param polling_interval: add message time gap to control chaos
    :param broker: backend (gcp-pubsub or redis)
    :param transform: message transformation
    :param schema: object
    :param topic_in: entry topic's name
    :param topic_out: exit topic's name
    :param subscription: subscription's name
    :return: None
    """
    thread = threading.Thread(target=consumeAndPublish,
                              args=(broker, transform, topic_in, topic_out,
                                    subscription, schema, polling_interval),
                              daemon=True)

    # publish and consume
    thread.start()

    return thread
